# Copyright 2021 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Component test for the sampling component."""

import tensorflow as tf
from absl.testing import absltest
from tfx.types import artifact_utils, channel_utils, standard_artifacts
from tfx.utils import json_utils

from tfx_addons.sampling import component, spec


class ComponentTest(absltest.TestCase):
  def testConstruct(self):
    examples = standard_artifacts.Examples()
    examples.split_names = artifact_utils.encode_split_names(['train', 'eval'])
    params = {
        spec.SAMPLER_INPUT_KEY: channel_utils.as_channel([examples]),
        spec.SAMPLER_SPLIT_KEY: ['train'],
        spec.SAMPLER_LABEL_KEY: 'label'
    }

    under = component.Sampler(**params)

    self.assertEqual(standard_artifacts.Examples.TYPE_NAME,
                     under.outputs[spec.SAMPLER_OUTPUT_KEY].type_name)
    self.assertEqual(under.spec.exec_properties[spec.SAMPLER_SPLIT_KEY],
                     json_utils.dumps(['train']))
    self.assertEqual(under.spec.exec_properties[spec.SAMPLER_LABEL_KEY],
                     'label')

  def testConstructWithOptions(self):
    examples = standard_artifacts.Examples()
    examples.split_names = artifact_utils.encode_split_names(['train', 'eval'])
    params = {
        spec.SAMPLER_INPUT_KEY: channel_utils.as_channel([examples]),
        spec.SAMPLER_LABEL_KEY: 'test_label',
        spec.SAMPLER_NAME_KEY: 'test_name',
        spec.SAMPLER_SPLIT_KEY: ['train', 'eval'],
        spec.SAMPLER_COPY_KEY: False,
        spec.SAMPLER_SHARDS_KEY: 10,
        spec.SAMPLER_CLASSES_KEY: ['label']
    }

    under = component.Sampler(**params)

    self.assertEqual(standard_artifacts.Examples.TYPE_NAME,
                     under.outputs[spec.SAMPLER_OUTPUT_KEY].type_name)
    self.assertEqual(under.spec.exec_properties[spec.SAMPLER_LABEL_KEY],
                     'test_label')
    self.assertEqual(under.spec.exec_properties[spec.SAMPLER_NAME_KEY],
                     'test_name')
    self.assertEqual(under.spec.exec_properties[spec.SAMPLER_SPLIT_KEY],
                     json_utils.dumps(['train', 'eval']))
    self.assertEqual(under.spec.exec_properties[spec.SAMPLER_COPY_KEY], False)
    self.assertEqual(under.spec.exec_properties[spec.SAMPLER_SHARDS_KEY], 10)
    self.assertEqual(under.spec.exec_properties[spec.SAMPLER_CLASSES_KEY],
                     json_utils.dumps(['label']))


if __name__ == '__main__':
  tf.test.main()



# Copyright 2021 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for schemaCuration.component.executor."""

import os

import pytest
import tensorflow as tf
from pkg_resources import resource_filename
from tfx import types
from tfx.dsl.io import fileio
from tfx.types import standard_artifacts, standard_component_specs

from tfx_addons.schema_curation.component import executor
from tfx_addons.schema_curation.test_data.module_file import module_file


# ToDo(casassg): Test fails currently. Marking as skip until it gets resolved so that
# CI is not red.
@pytest.mark.skip
class ExecutorTest(tf.test.TestCase):
  def testDo(self):
    super().setUp()

    self._output_data_dir = os.path.join(
        os.environ.get('TEST_UNDECLARED_OUTPUTS_DIR', self.get_temp_dir()),
        self._testMethodName)

    self.schema = standard_artifacts.Schema()
    schema_file = resource_filename(
        'tfx_addons.schema_curation.test_data.schema_gen', 'schema.pbtxt')
    fileio.copy(
        schema_file,
        os.path.join(self.get_temp_dir(), 'input_schema', 'schema.pbtxt'))
    self.schema.uri = os.path.join(self.get_temp_dir(), 'input_schema')

    self.input_dict = {
        standard_component_specs.SCHEMA_KEY: [self.schema],
    }

    self.schema_output = standard_artifacts.Schema()
    self.schema_output.uri = os.path.join(self._output_data_dir,
                                          'custom_schema')

    output_dict = {
        'custom_schema': [self.schema_output],
    }

    self._module_file = module_file.__file__
    self.schema_fn = '%s.%s' % (module_file.schema_fn.__module__,
                                module_file.schema_fn.__name__)

    print(self._module_file)
    self.exec_properties = {
        standard_component_specs.MODULE_FILE_KEY: self._module_file
    }

    self.schema_curation_executor = executor.Executor()

    self.schema_curation_executor.Do(self.input_dict, output_dict,
                                     self.exec_properties)

  def _verify_schema_curation_outputs(self):
    self.assertNotEqual(0, len(fileio.listdir(self.schema_output.uri)))

  def testDoWithModuleFile(self):
    self.exec_properties['module_file'] = self._module_file
    self.schema_curation_executor.Do(self.input_dict, self.output_dict,
                                     self.exec_properties)
    self._verify_schema_curation_outputs()

  def get_schema_fn(self):
    self._exec_properties['schema_fn'] = self.schema_fn
    self.schema_curation_executor.Do(self.input_dict, self.output_dict,
                                     self._exec_properties)
    self._verify_schema_curation_outputs()

  def testDoWithCache(self):
    # First run that creates cache.
    output_cache_artifact = types.Artifact('OutputCache')
    output_cache_artifact.uri = os.path.join(self._output_data_dir, 'CACHE/')

    self.output_dict['cache_output_path'] = [output_cache_artifact]

    self.exec_properties['module_file'] = self._module_file
    self.schemaCuration_executor.Do(self.input_dict, self.output_dict,
                                    self.exec_properties)
    self._verify_schema_curation_outputs()
    self.assertNotEqual(0, len(tf.io.gfile.listdir(output_cache_artifact.uri)))


if __name__ == '__main__':
  tf.test.main()



# Copyright 2021 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Executor for Sampler component."""

import os
import random
from typing import Any, Dict, List, Text

import apache_beam as beam
import tensorflow as tf
from tfx import types
from tfx.dsl.components.base import base_beam_executor
from tfx.dsl.io import fileio
from tfx.types import artifact_utils
from tfx.utils import io_utils, json_utils

from tfx_addons.sampling import spec


class Executor(base_beam_executor.BaseBeamExecutor):
  """Executor for Sampler."""
  def _CreatePipeline(self, unused_transform_output: Text) -> beam.Pipeline:
    """Creates beam pipeline.
    Args:
      unused_transform_output: unused.
    Returns:
      Beam pipeline.
    """

    return self._make_beam_pipeline()

  def Do(
      self,
      input_dict: Dict[Text, List[types.Artifact]],
      output_dict: Dict[Text, List[types.Artifact]],
      exec_properties: Dict[Text, Any],
  ) -> None:
    """Sampler executor entrypoint.
    Args:
      input_dict: Input dict from input key to a list of artifacts, including:
        - examples: A list of type `standard_artifacts.Examples` which should
          contain custom splits specified in splits_config. If custom split is
          not provided, this should contain two splits 'train' and 'eval'.
      output_dict: Output dict from key to a list of artifacts, including:
        - sampled_examples: sampled examples, only for the given
          splits as specified in splits. May also include copies of the
          other non-sampled spits, as specified by null_classes.
      exec_properties: A dict of execution properties, including:
        - name: Optional unique name. Necessary if multiple components are
          declared in the same pipeline.
        - label: The name of the column containing class names to sample by.
        - splits: A list containing splits to sample.
        - copy_others: Determines whether we copy over the splits that aren't
          sampled, or just exclude them from the output artifact. Defaults
          to True.
        - shards: The number of files that each sampled split should
          contain. Default 0 is Beam's tfrecordio function's default.
        - null_classes: A list determining which classes that we should
          not sample. Defaults to None.
        - sampling_strategy: An enum of type SamplingStrategy, determining if
          the executor should over or undersample.
    Returns:
      None
    """

    self._log_startup(input_dict, output_dict, exec_properties)

    label = exec_properties[spec.SAMPLER_LABEL_KEY]
    sampling_strategy = exec_properties[spec.SAMPLER_SAMPLE_KEY]
    splits = json_utils.loads(exec_properties[spec.SAMPLER_SPLIT_KEY])
    copy_others = exec_properties[spec.SAMPLER_COPY_KEY]
    shards = exec_properties[spec.SAMPLER_SHARDS_KEY]
    null_classes = json_utils.loads(exec_properties[spec.SAMPLER_CLASSES_KEY])

    input_artifact = artifact_utils.get_single_instance(
        input_dict[spec.SAMPLER_INPUT_KEY])
    output_artifact = artifact_utils.get_single_instance(
        output_dict[spec.SAMPLER_OUTPUT_KEY])

    if sampling_strategy not in (spec.SamplingStrategy.UNDERSAMPLE,
                                 spec.SamplingStrategy.OVERSAMPLE):
      raise ValueError("Invalid sampling strategy!")

    if not splits:
      raise ValueError("No splits are marked for sampling!")

    for split in splits:
      if split not in input_artifact.split_names:
        raise ValueError(
            f"Invalid split name {split} is not in input artifact!")

    if shards < 0:
      raise ValueError("Shards value must be non-negative!")

    if copy_others:
      output_artifact.split_names = input_artifact.split_names
    else:
      output_artifact.split_names = artifact_utils.encode_split_names(splits)

    # Fetch the input uri for each split
    split_data = {}
    for split in artifact_utils.decode_split_names(input_artifact.split_names):
      uri = artifact_utils.get_split_uri([input_artifact], split)
      split_data[split] = uri

    for split, uri in split_data.items():
      if split in splits:  # Undersampling split
        output_dir = artifact_utils.get_split_uri([output_artifact], split)
        split_dir = os.path.join(output_dir, f"Split-{split}")
        with self._CreatePipeline(split_dir) as p:
          data = read_tfexamples(p, uri, label)
          merged = sample_examples(data, null_classes, sampling_strategy)
          write_tfexamples(merged, shards, split_dir)
      elif copy_others:  # Copy the other split if copy_others is True
        input_dir = uri
        output_dir = artifact_utils.get_split_uri([output_artifact], split)
        for filename in fileio.listdir(input_dir):
          input_uri = os.path.join(input_dir, filename)
          output_uri = os.path.join(output_dir, filename)
          io_utils.copy_file(src=input_uri, dst=output_uri, overwrite=True)


def _generate_elements(example, label):
  """Function that fetches the class label from a tf.Example and returns one
  item in a K-V PCollection with the key as the label and the value as the
  string-parsed tf.Example.

  Args:
    example: a tf.Example in serialized format, taken directly from a
      TFRecordDataset.
    label: string containing the name of the categorical variable that we are
      extracting from the example.
  Returns:
    Tuple with two items. First item is a class label; second item is the input
      tf.Example, deserialized and parsed from string format.
  """

  class_label = None
  parsed = tf.train.Example.FromString(example.numpy())
  if parsed.features.feature[label].int64_list.value:
    val = parsed.features.feature[label].int64_list.value
    if len(val) > 0:
      class_label = val[0]
  else:
    val = parsed.features.feature[label].bytes_list.value
    if len(val) > 0:
      class_label = val[0].decode()
  return (class_label, parsed)


def sample_data(_,
                val,
                sampling_strategy=spec.SamplingStrategy.UNDERSAMPLE,
                side=0):
  """Function called in a Beam pipeline that performs sampling using Python's
  random module on an input key:value pair, where the key is the class label
  and the values are the data points to sample. Note that the key is discarded."""

  if sampling_strategy == spec.SamplingStrategy.UNDERSAMPLE:
    random_sample_data = random.sample(val, side)
  elif sampling_strategy == spec.SamplingStrategy.OVERSAMPLE:
    random_sample_data = random.choices(val, k=side)
  else:
    raise ValueError("Invalid value for sampling_strategy variable!")

  for item in random_sample_data:
    yield item


def filter_null(item, keep_null=False, null_vals=None):
  """Function that returns or doesn't return the inputted item if its first
  value is either a False value or is in the inputted null_vals list.

  This function first determines if the item's first value is equivalent to
  False using bool(), with one exception; 0 is considered as "True". If the
  first value is in null_vals, the first value is automatically considered a
  "null value" and is therefore considered to be False. If the value is False,
  then None is returned; if the value is True, then the original item is
  returned. The keep_null value reverses this, so True values return None,
  and False values return the item.

  Args:
    item: Tuple whose first value determines whether it is returned or not.
      Should always be a two-value tuple, with the first value being the class
      value and the second value being all examples that belong to that class.
    keep_null: Determines whether we keep False/"null" values or True/not
      "null" values.
    null_vals: List containing values that should be considered as False/"null".
  Returns:
    None or the inputted item, depending on if the item is False/in null_vals,
      and then depending on the value of keep_null.
  """

  if item[0] == 0:
    keep = True
  else:
    keep = bool(item[0])

  if null_vals and str(item[0]) in null_vals and keep:
    keep = False
  keep ^= keep_null

  return item if keep else None


def read_tfexamples(p, uri, label):
  """Function that reads tf.Examples from tfRecord files and converts them
  to a K-V PCollection usable by Beam."""

  dataset = tf.data.TFRecordDataset(tf.data.Dataset.list_files(f'{uri}/*'),
                                    compression_type="GZIP")

  # Take the input TFRecordDataset and extract the class label that we want.
  # Output format is a K-V PCollection: {class_label: TFRecord in string format}
  data = (p
          | "DatasetToPCollection" >> beam.Create(dataset)
          | "MapToLabel" >> beam.Map(_generate_elements, label))
  return data


def sample_examples(data, null_classes, sampling_strategy):
  """Function that performs the sampling given a label-mapped dataset."""

  # Finds the minimum frequency of all classes in the input label.
  # Output is a singleton PCollection with the minimum # of examples.

  def find_minimum(elements):
    return min(elements or [0])

  def find_maximum(elements):
    return max(elements or [0])

  if sampling_strategy == spec.SamplingStrategy.UNDERSAMPLE:
    sample_fn = find_minimum
  elif sampling_strategy == spec.SamplingStrategy.OVERSAMPLE:
    sample_fn = find_maximum
  else:
    raise ValueError("Invalid value for sampling_strategy variable!")

  val = (data
         | "CountPerKey" >> beam.combiners.Count.PerKey()
         | "FilterNullCount" >>
         beam.Filter(lambda x: filter_null(x, null_vals=null_classes))
         | "Values" >> beam.Values()
         | "GetSample" >> beam.CombineGlobally(sample_fn))

  # Actually performs the undersampling functionality.
  # Output format is a K-V PCollection: {class_label: TFRecord in string format}
  res = (data
         | "GroupBylabel" >> beam.GroupByKey()
         | "FilterNull" >>
         beam.Filter(lambda x: filter_null(x, null_vals=null_classes))
         | "Sample" >> beam.FlatMapTuple(sample_data,
                                         sampling_strategy=sampling_strategy,
                                         side=beam.pvalue.AsSingleton(val)))

  # Take out all the null values from the beginning and put them back in the pipeline
  null = (data
          | "ExtractNull" >> beam.Filter(
              lambda x: filter_null(x, keep_null=True, null_vals=null_classes))
          | "NullValues" >> beam.Values())
  return (res, null) | "Merge PCollections" >> beam.Flatten()


def write_tfexamples(examples, shards, output_dir):
  # Write the final set of TFRecords to the output artifact's files.
  _ = (examples
       | "Serialize" >> beam.Map(lambda x: x.SerializeToString())
       | "WriteToTFRecord" >> beam.io.tfrecordio.WriteToTFRecord(
           output_dir,
           file_name_suffix=".gz",
           num_shards=shards,
           compression_type=beam.io.filesystem.CompressionTypes.GZIP,
       ))



# Copyright 2021 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""TFX Schema curation custom component executor"""

import os
from typing import Any, Dict, List, Text

from absl import logging
from tfx import types
from tfx.components.util import udf_utils
from tfx.dsl.components.base import base_executor
from tfx.types import artifact_utils, standard_component_specs
from tfx.utils import io_utils

_DEFAULT_FILE_NAME = "schema.pbtxt"


class Executor(base_executor.BaseExecutor):
  """Executor for TFX Schema Curation Custom Component.

    This Executor will execute the `schema_fn` with correct parameters
    by resolving the input artifacts, output artifacts and execution properties.
    """
  def Do(
      self,
      input_dict: Dict[Text, List[types.Artifact]],
      output_dict: Dict[Text, List[types.Artifact]],
      exec_properties: Dict[Text, Any],
  ) -> None:
    """Uses a user-supplied `schema_fn` to curate a schema.

        The schema curation Executor invokes a `schema_fn` callback function
        provided by the user via the `module_file` parameter.
        In this function, user changes the schema.

        `schema_fn` should be written in the form:
        def schema_fun (Schema) -> Schema:
           #code to modify schema

        Args:
          input_dict: Input dict from input key to a list of ML-Metadata
            Artifacts including:
            - schema: Schema of the data. A list of type
              `standard_artifacts.Schema` which should
              contain a single schema artifact.
          output_dict: Output dict from output key to a list of Artifacts.
            - schema: Schema of the data. A list of type
              `standard_artifacts.Schema` which should contain
               a single schema artifact.
          exec_properties: A dict of execution properties, including:
            - module_file: The file path to a python module file, from which the
              `schema_fn` function will be loaded. Exactly one of
              `module_file`, `module_path` and `schema_fn` should be set.
            - module_path: The python module path, from which the
              `schema_fn` function will be loaded. Exactly one of
              `module_file`, `module_path` and `schema_fn` should be set.
            - schema_fn: The module path to a python function that
              implements `schema_fn`. Exactly one of `module_file`,
              `module_path` and `schema_fn` should be set.

        Returns:
          None

        Raises:
          None
        """
    self._log_startup(input_dict, output_dict, exec_properties)

    # get uri of schema artifact
    schema_path = io_utils.get_only_uri_in_dir(
        artifact_utils.get_single_uri(
            input_dict.get(standard_component_specs.SCHEMA_KEY)))

    schema = io_utils.SchemaReader().read(schema_path)

    # load and run user defined `schema_fn` from module file given by user
    schema_fn = udf_utils.get_fn(exec_properties, 'schema_fn')
    custom_schema = schema_fn(schema)

    # Write custom schema back into the pipeline
    output_uri = os.path.join(
        artifact_utils.get_single_uri(output_dict['custom_schema']),
        _DEFAULT_FILE_NAME)

    io_utils.write_pbtxt_file(output_uri, custom_schema)
    logging.info("Schema written to %s.", output_uri)



# Copyright 2021 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""
Generic TFX FeastExampleGen executor.
"""

import os
from typing import Any, Dict, Optional

import apache_beam as beam
import feast
from absl import logging
from apache_beam.options import value_provider
from feast.infra.offline_stores.bigquery import BigQueryRetrievalJob
from feast.infra.offline_stores.offline_store import RetrievalJob
from google.protobuf import json_format
from google.protobuf.json_format import MessageToDict
from google.protobuf.struct_pb2 import Struct
from tfx.components.example_gen import base_example_gen_executor
from tfx.extensions.google_cloud_big_query import utils
from tfx.proto import example_gen_pb2

from tfx_addons.feast_examplegen import converters

_REPO_CONFIG_KEY = "repo_conf"
_FEATURE_KEY = "feature_refs"
_FEATURE_SERVICE_KEY = "feature_service_ref"


def _load_custom_config(custom_config):
  config = example_gen_pb2.CustomConfig()
  json_format.Parse(custom_config, config)
  s = Struct()
  config.custom_config.Unpack(s)
  config_dict = MessageToDict(s)
  return config_dict


def _load_feast_feature_store(
    custom_config: Dict[str, Any]) -> feast.FeatureStore:
  repo_config = feast.RepoConfig.parse_raw(custom_config[_REPO_CONFIG_KEY])
  return feast.FeatureStore(config=repo_config)


def _get_retrieval_job(entity_query: str,
                       custom_config: Dict[str, Any]) -> RetrievalJob:
  """Get feast retrieval job

    Args:
        entity_query (str): entity query.
        custom_config (Dict[str, Any]): Custom configuration from component

    Raises:
        RuntimeError: [description]

    Returns:
        RetrievalJob: [description]
    """
  feature_list = custom_config.get(_FEATURE_KEY, None)
  feature_service = custom_config.get(_FEATURE_SERVICE_KEY, None)
  fs = _load_feast_feature_store(custom_config)

  if feature_list:
    features = feature_list
  elif feature_service:
    features = fs.get_feature_service(feature_service)
  else:
    raise RuntimeError(
        "Either feature service or feature list should be provided")

  return fs.get_historical_features(entity_df=entity_query, features=features)


def _get_gcp_project(exec_properties: Dict[str, Any]) -> Optional[str]:
  # Get GCP project from exec_properties.
  beam_pipeline_args = exec_properties["_beam_pipeline_args"]
  pipeline_options = beam.options.pipeline_options.PipelineOptions(
      beam_pipeline_args)

  # Try to parse the GCP project ID from the beam pipeline options.
  project = pipeline_options.view_as(
      beam.options.pipeline_options.GoogleCloudOptions).project
  if isinstance(project, value_provider.ValueProvider):
    return project.get()
  return project or None


def _get_datasource_converter(exec_properties: Dict[str, Any],
                              split_pattern: str):
  # Load custom config dictionary
  custom_config = _load_custom_config(exec_properties["custom_config"])

  # Get Feast retrieval job
  retrieval_job = _get_retrieval_job(entity_query=split_pattern,
                                     custom_config=custom_config)

  # Setup datasource and converter.
  if isinstance(retrieval_job, BigQueryRetrievalJob):
    table = retrieval_job.to_bigquery()
    query = f'SELECT * FROM {table}'
    # Internally Beam creates a temporary table and exports from the query.
    datasource = utils.ReadFromBigQuery(query=query)
    converter = converters._BigQueryConverter(  # pylint: disable=protected-access
        query, _get_gcp_project(exec_properties))
  else:
    raise NotImplementedError(
        f"Support for {type(retrieval_job)} is not available yet. "
        "For now we only support BigQuery source.")

  return datasource, converter


@beam.ptransform_fn
@beam.typehints.with_input_types(beam.Pipeline)
@beam.typehints.with_output_types(bytes)
def _FeastToExampleTransform(  # pylint: disable=invalid-name
    pipeline: beam.Pipeline, exec_properties: Dict[str, Any],
    split_pattern: str) -> beam.pvalue.PCollection:
  """Read from BigQuery and transform to TF examples.

    Args:
      pipeline: beam pipeline.
      exec_properties: A dict of execution properties.
      split_pattern: Split.pattern in Input config, a BigQuery sql string.

    Returns:
      PCollection of TF examples.
    """
  # Feast doesn't allow us to configure GCP project used to explore BQ (afaik),
  # we instead set the beam project as default in environment as sometimes
  # it may not work on the default project
  gcp_project = _get_gcp_project(exec_properties)
  logging.info("Detected GCP project %s", gcp_project)
  restore_project = None
  if gcp_project:
    logging.info("Overwriting GOOGLE_CLOUD_PROJECT env var to %s", gcp_project)
    restore_project = os.environ.get("GOOGLE_CLOUD_PROJECT", None)
    os.environ["GOOGLE_CLOUD_PROJECT"] = gcp_project
  try:
    datasource, converter = _get_datasource_converter(
        exec_properties=exec_properties, split_pattern=split_pattern)
  finally:
    if restore_project:
      os.environ["GOOGLE_CLOUD_PROJECT"] = restore_project

  # Setup converter from dictionary of str -> value to bytes
  map_function = None
  out_format = exec_properties.get("output_data_format",
                                   example_gen_pb2.FORMAT_TF_EXAMPLE)
  if out_format == example_gen_pb2.FORMAT_TF_EXAMPLE:
    map_function = converter.RowToExampleBytes
  elif out_format == example_gen_pb2.FORMAT_TF_SEQUENCE_EXAMPLE:
    map_function = converter.RowToSequenceExampleBytes
  else:
    raise NotImplementedError(
        f"Format {out_format} is not currently supported."
        " Currently we only support tfexample")

  # Setup pipeline
  return (pipeline
          | "DataRetrieval" >> datasource
          | "ToBytes" >> beam.Map(map_function))


class Executor(base_example_gen_executor.BaseExampleGenExecutor):
  """Generic TFX FeastExampleGen executor."""
  def GetInputSourceToExamplePTransform(self) -> beam.PTransform:
    """Returns PTransform for Feast to bytes."""
    return _FeastToExampleTransform



# Copyright 2021 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Sampling component definition."""

import enum
from typing import Text

from tfx import types
from tfx.types import standard_artifacts
from tfx.types.component_spec import ChannelParameter, ExecutionParameter

SAMPLER_INPUT_KEY = 'input_data'
SAMPLER_OUTPUT_KEY = 'output_data'
SAMPLER_LABEL_KEY = 'label'
SAMPLER_NAME_KEY = 'name'
SAMPLER_SPLIT_KEY = 'splits'
SAMPLER_COPY_KEY = 'copy_others'
SAMPLER_SHARDS_KEY = 'shards'
SAMPLER_CLASSES_KEY = 'null_classes'
SAMPLER_SAMPLE_KEY = 'sampling_strategy'


class SamplingStrategy(enum.IntEnum):
  """Determines which kind of sampling to perform."""
  UNDERSAMPLE = 1
  OVERSAMPLE = 2


class SamplerSpec(types.ComponentSpec):
  """Sampling component spec."""

  PARAMETERS = {
      SAMPLER_LABEL_KEY: ExecutionParameter(type=str),
      SAMPLER_NAME_KEY: ExecutionParameter(type=Text, optional=True),
      SAMPLER_SPLIT_KEY: ExecutionParameter(type=str, optional=True),
      SAMPLER_COPY_KEY: ExecutionParameter(type=int, optional=True),
      SAMPLER_SHARDS_KEY: ExecutionParameter(type=int, optional=True),
      SAMPLER_CLASSES_KEY: ExecutionParameter(type=str, optional=True),
      SAMPLER_SAMPLE_KEY: ExecutionParameter(type=int, optional=True)
  }
  INPUTS = {
      SAMPLER_INPUT_KEY: ChannelParameter(type=standard_artifacts.Examples),
  }
  OUTPUTS = {
      SAMPLER_OUTPUT_KEY: ChannelParameter(type=standard_artifacts.Examples),
  }



# Copyright 2022 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for tfx_addons.message_exit_handler.component."""

import json
import unittest.mock as mock

import pytest
import tensorflow as tf
from tfx import v1 as tfx

from tfx_addons.message_exit_handler import constants
from tfx_addons.message_exit_handler.message_providers import base_provider
from tfx_addons.utils.test_utils import get_tfx_version


def fake_decryption_fn(encrypted_message):
  return encrypted_message.upper()


class ComponentTest(tf.test.TestCase):
  @staticmethod
  def get_final_status(state: str = constants.SUCCESS_STATUS,
                       error: str = "") -> str:
    """Assemble final status for tests"""
    # final status proto
    status = {
        "state":
        state,
        "error":
        error,
        "pipelineJobResourceName":
        ("projects/test-project/locations/us-central1/"
         "pipelineJobs/test-pipeline-job"),
    }
    if error:
      status.update({"error": {"message": error}})
    return json.dumps(status)

  @pytest.mark.skipif(get_tfx_version(tfx.__version__) < (1, 6, 0),
                      reason="not supported version")
  @mock.patch("tfx.orchestration.kubeflow.v2.decorators.exit_handler",
              lambda x: x)
  def test_component_fn(self):

    # import in function to pass TFX 1.4 tests
    from tfx_addons.message_exit_handler import \
        component  # pylint: disable=import-outside-toplevel
    final_status = self.get_final_status()

    with self.assertLogs(level="INFO") as logs:
      component.MessageExitHandler(final_status=final_status,
                                   on_failure_only=True)

      self.assertLen(logs.output, 1)
      self.assertEqual(
          "INFO:absl:MessageExitHandler: Skipping notification on success.",
          logs.output[0],
      )

  @pytest.mark.skipif(get_tfx_version(tfx.__version__) < (1, 6, 0),
                      reason="not supported version")
  @mock.patch(
      "tfx_addons.message_exit_handler.message_providers.slack_provider.WebClient"  # pylint: disable=line-too-long
  )
  @mock.patch("tfx.orchestration.kubeflow.v2.decorators.exit_handler",
              lambda x: x)
  def test_component_slack(self, mock_web_client):

    # import in function to pass TFX 1.4 tests
    from tfx_addons.message_exit_handler import \
        component  # pylint: disable=import-outside-toplevel
    final_status = self.get_final_status()
    creds = json.dumps({"slack_token": "token", "slack_channel_id": "channel"})

    with self.assertLogs(level="INFO"):
      component.MessageExitHandler(
          final_status=final_status,
          message_type=base_provider.MessagingType.SLACK.value,
          slack_credentials=creds,
      )

      mock_web_client.assert_called_once()
      mock_web_client.assert_called_with(token="token")

  @pytest.mark.skipif(get_tfx_version(tfx.__version__) < (1, 6, 0),
                      reason="not supported version")
  @mock.patch(
      "tfx_addons.message_exit_handler.message_providers.slack_provider.WebClient"  # pylint: disable=line-too-long
  )
  @mock.patch("tfx.orchestration.kubeflow.v2.decorators.exit_handler",
              lambda x: x)
  def test_component_slack_decrypt(self, mock_web_client):

    # import in function to pass TFX 1.4 tests
    from tfx_addons.message_exit_handler import \
        component  # pylint: disable=import-outside-toplevel
    final_status = self.get_final_status()
    creds = json.dumps({"slack_token": "token", "slack_channel_id": "channel"})

    with self.assertLogs(level="INFO"):
      component.MessageExitHandler(
          final_status=final_status,
          message_type=base_provider.MessagingType.SLACK.value,
          slack_credentials=creds,
          decrypt_fn=
          "tfx_addons.message_exit_handler.component_test.fake_decryption_fn",
      )

      mock_web_client.assert_called_once()
      mock_web_client.assert_called_with(token="TOKEN")


if __name__ == "__main__":
  tf.test.main()



# Copyright 2021 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""MLMD representations for TFX Pipeline. References objects from MLMD to inspect their status.
"""

from functools import lru_cache
from typing import List, Optional

from ml_metadata.metadata_store import MetadataStore
from ml_metadata.proto import Artifact, metadata_store_pb2
from tfx.orchestration.metadata import _CONTEXT_TYPE_PIPELINE
from tfx.orchestration.pipeline import Pipeline


class _MetadataClientBase:
  def __init__(
      self, metadata_connection_config: metadata_store_pb2.ConnectionConfig):
    self._metadata_connection_config = metadata_connection_config
    self._store = None

  @property
  def _mlmd(self):
    if not self._store:
      self._store = MetadataStore(self._metadata_connection_config)
    return self._store

  @property
  def connection_config(self):
    return self._metadata_connection_config

  @property
  @lru_cache(None)
  def artifact_types(self):
    return {x.name: int(x.id) for x in self._mlmd.get_artifact_types()}


class ModelArtifact(_MetadataClientBase):
  """Context for a specific synchronous pipeline run.
    Unique pipeline_name + run_id.
    """
  def __init__(
      self,
      model_artifact: metadata_store_pb2.Artifact,
      metadata_connection_config: metadata_store_pb2.ConnectionConfig,
  ) -> None:
    super().__init__(metadata_connection_config)
    self._artifact = model_artifact

  def __repr__(self) -> str:
    return f'{self.__class__.__name__}(pipeline_name="{self.pipeline_name}",'\
    ' run_id="{self.run_id}"")'


class PipelineContext(_MetadataClientBase):
  """Reference to pipeline context in MLMD."""
  def __init__(
      self,
      pipeline_context: metadata_store_pb2.Context,
      metadata_connection_config: metadata_store_pb2.ConnectionConfig,
  ) -> None:
    """Creates instance of PipelineContext
        Args:
            pipeline_context: MLMD Context for a pipeline.
            metadata_connection_config: MLMD connection config.
        """
    super().__init__(metadata_connection_config)
    self._context = pipeline_context

  @staticmethod
  def get_pipeline_context(mlmd_client: MetadataStore,
                           pipeline_name: str) -> metadata_store_pb2.Context:
    """Static method to get pipeline context.
        Args:
            mlmd_client: MLMD MetadataStore instance
            pipeline_name: Pipeline name to retrieve
        Returns:
            metadata_store_pb2.Context: Returns None if context doesn't
                exist in MLMD yet. This can happen if pipeline has not
                been executed yet.
        """
    return mlmd_client.get_context_by_type_and_name(_CONTEXT_TYPE_PIPELINE,
                                                    pipeline_name)

  def get_artifact_by_type_name(self, type_name: str) -> List[Artifact]:
    """Returns artifacts of a given type.

      Args:
        type_name: Name of artifact type to retrieve.
      Returns:
        List[Artifact]: List of artifacts of given type.
    """
    artifacts = self._mlmd.get_artifacts_by_context(self._context.id)
    return [
        artifact for artifact in artifacts
        if artifact.type_id == self.artifact_types.get(type_name, -1)
    ]

  @property
  def name(self) -> str:
    return self._context.name

  def __repr__(self) -> str:
    return f'{self.__class__.__name__}(name="{self.name}")'


class MetadataClient(_MetadataClientBase):
  """MLMD read client."""
  def __init__(
      self,
      metadata_connection_config: metadata_store_pb2.ConnectionConfig = None,
  ):
    """Creates instance of MetadataClient.
        Args:
            metadata_config: Configuration to connect to MLMD.
                Defaults to staging DB connection.
        """
    super().__init__(metadata_connection_config)

  @staticmethod
  def from_pipeline(pipeline: Pipeline) -> Optional[PipelineContext]:
    """Retrieve PipelineContext given a pipeline.
        Helper method to make this recurring task easier.

        Args:
            pipeline: Pipeline to retrieve context from.
        Returns:
            Optional[PipelineContext]: Returns None if context doesn't
                exist in MLMD yet. This can happen if pipeline has not been
                executed yet.
        """
    return MetadataClient(pipeline.metadata_connection_config).get_pipeline(
        pipeline.pipeline_info.pipeline_name)

  def get_pipeline(self, pipeline_name: str) -> Optional[PipelineContext]:
    """Retrieve PipelineContext given a pipeline name.
        Args:
            pipeline_name: Name of pipeline to retrieve
        Returns:
            Optional[PipelineContext]: Returns None if context
                doesn't exist in MLMD yet. This can happen if pipeline has
                not been executed yet.
        """
    context = PipelineContext.get_pipeline_context(self._mlmd, pipeline_name)
    if not context:
      return None
    return PipelineContext(context, self.connection_config)

  def __repr__(self) -> str:
    connection_str = str(self.connection_config).replace("\n", ",")
    return f"{self.__class__.__name__}(connection_config={{{connection_str}}})"



# Copyright 2021 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Data converter library to convert from source data to serialized data to be stored out."""
import abc
from typing import Any, Dict, Optional

import tensorflow as tf
from google.cloud import bigquery


class _Converter(abc.ABC):
  """Takes in data as a dictionary of string -> value and returns the serialized form
    of whatever representation we want.
    """
  @abc.abstractmethod
  def RowToExampleBytes(self, instance: Any) -> bytes:
    """Generate tf.Example bytes from dictionary.

        Args:
            instance (Any): Data row generated from data source

        Returns:
            bytes: Serialized tf.SequenceExample
        """
    pass

  @abc.abstractmethod
  def RowToSequenceExampleBytes(self, instance: Any) -> bytes:
    """Generate tf.SequenceExample bytes from dictionary.

        Args:
            instance (Any): Data row generated from data source

        Returns:
            bytes: Serialized tf.SequenceExample
        """
    pass


# NB(gcasassaez): Fork from tfx.extensions.google_cloud_big_query.utils
# This is mostly to add support for timestamp since its a common
# format used by Feast.
def row_to_example(  # pylint: disable=invalid-name
    field_to_type: Dict[str, str],
    field_name_to_data: Dict[str, Any]) -> tf.train.Example:
  """Convert bigquery result row to tf example.

  Args:
    field_to_type: The name of the field to its type from BigQuery.
    field_name_to_data: The data need to be converted from BigQuery that
      contains field name and data.

  Returns:
    A tf.train.Example that converted from the BigQuery row. Note that BOOLEAN
    type in BigQuery result will be converted to int in tf.train.Example.

  Raises:
    RuntimeError: If the data type is not supported to be converted.
      Only INTEGER, BOOLEAN, FLOAT, STRING is supported now.
  """
  feature = {}
  for key, value in field_name_to_data.items():
    data_type = field_to_type[key]

    if value is None:
      feature[key] = tf.train.Feature()
      continue

    value_list = value if isinstance(value, list) else [value]
    if data_type in ('INTEGER', 'BOOLEAN'):
      feature[key] = tf.train.Feature(int64_list=tf.train.Int64List(
          value=value_list))
    elif data_type == 'FLOAT':
      feature[key] = tf.train.Feature(float_list=tf.train.FloatList(
          value=value_list))
    elif data_type == 'TIMESTAMP':
      feature[key] = tf.train.Feature(float_list=tf.train.FloatList(
          value=[elem.timestamp() for elem in value_list]))
    elif data_type == 'STRING':
      feature[key] = tf.train.Feature(bytes_list=tf.train.BytesList(
          value=[tf.compat.as_bytes(elem) for elem in value_list]))
    else:
      # TODO(jyzhao): support more types.
      raise RuntimeError(
          'BigQuery column type {} is not supported.'.format(data_type))

  return tf.train.Example(features=tf.train.Features(feature=feature))


class _BigQueryConverter(_Converter):
  """Converter class for BigQuery source data"""
  def __init__(self, query: str, project: Optional[str]) -> None:
    client = bigquery.Client(project=project)
    # Dummy query to get the type information for each field.
    query_job = client.query("SELECT * FROM ({}) LIMIT 0".format(query))
    results = query_job.result()
    self._type_map = {}
    for field in results.schema:
      self._type_map[field.name] = field.field_type

  def RowToExampleBytes(self, instance: Dict[str, Any]) -> bytes:
    """Convert bigquery result row to tf example."""
    ex_pb2 = row_to_example(self._type_map, instance)
    return ex_pb2.SerializeToString(deterministic=True)

  def RowToSequenceExampleBytes(self, instance: Dict[str, Any]) -> bytes:
    """Convert bigquery result row to tf sequence example."""
    raise NotImplementedError("SequenceExample not implemented yet.")



# Copyright 2021 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Sampling pipeline example using TFX."""

from __future__ import absolute_import, division, print_function

import os
from typing import List, Text

import absl
import tensorflow_model_analysis as tfma
from tfx.components import (CsvExampleGen, Evaluator, ExampleValidator, Pusher,
                            SchemaGen, StatisticsGen, Trainer, Transform)
from tfx.components.trainer.executor import Executor
from tfx.dsl.components.base import executor_spec
from tfx.dsl.components.common import resolver
from tfx.dsl.experimental import (latest_artifacts_resolver,
                                  latest_blessed_model_resolver)
from tfx.orchestration import metadata, pipeline
from tfx.orchestration.local.local_dag_runner import LocalDagRunner
from tfx.proto import pusher_pb2, trainer_pb2
from tfx.types import Channel
from tfx.types.standard_artifacts import Model, ModelBlessing

from tfx_addons.sampling.component import Sampler

_pipeline_name = 'sampling_credit_card'
_sampling_root = os.path.dirname(__file__)
_data_root = os.path.join(_sampling_root, 'data')
# Python module file to inject customized logic into the TFX components. The
# Transform and Trainer both require user-defined functions to run successfully.
_module_file = os.path.join(_sampling_root, 'sampler_utils.py')
_serving_model_dir = os.path.join(_sampling_root, 'serving_model',
                                  _pipeline_name)
_tfx_root = os.path.join(os.environ['HOME'], 'tfx')
_pipeline_root = os.path.join(_tfx_root, 'pipelines', _pipeline_name)
_metadata_path = os.path.join(_tfx_root, 'metadata', _pipeline_name,
                              'metadata.db')

# Pipeline arguments for Beam powered Components.
_beam_pipeline_args = [
    '--direct_running_mode=multi_processing',
    # 0 means auto-detect based on on the number of CPUs available
    # during execution time.
    '--direct_num_workers=0',
]


def _create_pipeline(pipeline_name: Text, pipeline_root: Text, data_root: Text,
                     module_file: Text, serving_model_dir: Text,
                     metadata_path: Text,
                     beam_pipeline_args: List[Text]) -> pipeline.Pipeline:
  """Implements an example pipeline with the sampling component witin TFX."""

  example_gen = CsvExampleGen(input_base=data_root)
  statistics_gen = StatisticsGen(examples=example_gen.outputs['examples'])
  schema_gen = SchemaGen(statistics=statistics_gen.outputs['statistics'],
                         infer_feature_shape=False)
  example_validator = ExampleValidator(
      statistics=statistics_gen.outputs['statistics'],
      schema=schema_gen.outputs['schema'])

  # Sampler component, with input examples and class label.
  sample = Sampler(input_data=example_gen.outputs['examples'],
                   splits=['train'],
                   label='Class',
                   shards=10)

  transform = Transform(examples=sample.outputs['output_data'],
                        schema=schema_gen.outputs['schema'],
                        module_file=module_file)
  latest_model_resolver = resolver.Resolver(
      strategy_class=latest_artifacts_resolver.LatestArtifactsResolver,
      latest_model=Channel(type=Model)).with_id('latest_model_resolver')
  trainer = Trainer(
      module_file=module_file,
      custom_executor_spec=executor_spec.ExecutorClassSpec(Executor),
      transformed_examples=transform.outputs['transformed_examples'],
      schema=schema_gen.outputs['schema'],
      base_model=latest_model_resolver.outputs['latest_model'],
      transform_graph=transform.outputs['transform_graph'],
      train_args=trainer_pb2.TrainArgs(num_steps=10000),
      eval_args=trainer_pb2.EvalArgs(num_steps=5000))
  model_resolver = resolver.Resolver(
      strategy_class=latest_blessed_model_resolver.LatestBlessedModelResolver,
      model=Channel(type=Model),
      model_blessing=Channel(
          type=ModelBlessing)).with_id('latest_blessed_model_resolver')

  # Uses TFMA to compute a evaluation statistics over features of a model and
  # perform quality validation of a candidate model (compared to a baseline).
  eval_config = tfma.EvalConfig(
      model_specs=[tfma.ModelSpec(signature_name='eval')],
      slicing_specs=[
          tfma.SlicingSpec(),
          tfma.SlicingSpec(feature_keys=['trip_start_hour'])
      ],
      metrics_specs=[
          tfma.MetricsSpec(
              thresholds={
                  'accuracy':
                  tfma.config.MetricThreshold(
                      value_threshold=tfma.GenericValueThreshold(
                          lower_bound={'value': 0.6}),
                      change_threshold=tfma.GenericChangeThreshold(
                          direction=tfma.MetricDirection.HIGHER_IS_BETTER,
                          absolute={'value': -1e-10}))
              })
      ])

  evaluator = Evaluator(examples=example_gen.outputs['examples'],
                        model=trainer.outputs['model'],
                        baseline_model=model_resolver.outputs['model'],
                        eval_config=eval_config)
  pusher = Pusher(model=trainer.outputs['model'],
                  model_blessing=evaluator.outputs['blessing'],
                  push_destination=pusher_pb2.PushDestination(
                      filesystem=pusher_pb2.PushDestination.Filesystem(
                          base_directory=serving_model_dir)))

  return pipeline.Pipeline(
      pipeline_name=pipeline_name,
      pipeline_root=pipeline_root,
      components=[
          example_gen,
          statistics_gen,
          schema_gen,
          example_validator,
          sample,
          transform,
          latest_model_resolver,
          trainer,
          model_resolver,
          evaluator,
          pusher,
      ],
      enable_cache=False,
      metadata_connection_config=metadata.sqlite_metadata_connection_config(
          metadata_path),
      beam_pipeline_args=beam_pipeline_args)


# To run this pipeline from the python CLI:
#   $python sampler_pipeline_local.py
if __name__ == '__main__':
  absl.logging.set_verbosity(absl.logging.INFO)

  LocalDagRunner().run(
      _create_pipeline(pipeline_name=_pipeline_name,
                       pipeline_root=_pipeline_root,
                       data_root=_data_root,
                       module_file=_module_file,
                       serving_model_dir=_serving_model_dir,
                       metadata_path=_metadata_path,
                       beam_pipeline_args=_beam_pipeline_args))



# Copyright 2021 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""TFX Schema Curation Custom Component
"""

from typing import Optional, Text, Union

from tfx import types
from tfx.dsl.components.base import base_component, executor_spec
from tfx.orchestration import data_types
from tfx.types import standard_artifacts
from tfx.types.component_spec import ChannelParameter, ExecutionParameter

from tfx_addons.schema_curation.component import executor


class SchemaCurationSpec(types.ComponentSpec):
  """ComponentSpec for TFX Schema Curation Custom Component."""

  PARAMETERS = {
      'module_file': ExecutionParameter(type=str, optional=True),
      'module_path': ExecutionParameter(type=str, optional=True),
      'schema_fn': ExecutionParameter(type=str, optional=True)
  }
  INPUTS = {
      'schema':
      ChannelParameter(type=standard_artifacts.Schema
                       ),  # Dictionary obtained as output from SchemaGen
  }
  OUTPUTS = {
      'custom_schema':
      ChannelParameter(type=standard_artifacts.Schema
                       )  # Dictionary which containes new schema
  }


class SchemaCuration(base_component.BaseComponent):
  """Custom TFX Schema Curation Component.

    The SchemaCuration component is used to apply user code to a schema
    generated by SchemaGen in order to curate the schema based on
    domain knowledge.

    Component `outputs` contains:
     - `custom_schema`: Channel of type `standard_artifact.Schema`
    """

  SPEC_CLASS = SchemaCurationSpec
  EXECUTOR_SPEC = executor_spec.ExecutorClassSpec(executor.Executor)

  def __init__(
      self,
      schema: types.Channel,
      module_file: Optional[Union[Text, data_types.RuntimeParameter]] = None,
      module_path: Optional[Union[Text, data_types.RuntimeParameter]] = None,
      schema_fn: Optional[Union[Text, data_types.RuntimeParameter]] = None):
    """Construct a SchemaCurationComponent.

        Args:
          schema: A dictionary that containes the schema generated by
            SchemaGen component of tfx
          custom_schema: A dictionary that contains the schema after curation
            by the custom schema curation component
        """

    custom_schema = types.Channel(type=standard_artifacts.Schema)

    spec = SchemaCurationSpec(schema=schema,
                              custom_schema=custom_schema,
                              module_file=module_file,
                              module_path=module_path,
                              schema_fn=schema_fn)
    super().__init__(spec=spec)



# Copyright 2022 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Define TFX Addons version information."""

# We follow Semantic Versioning (https://semver.org/)
_MAJOR_VERSION = "0"
_MINOR_VERSION = "2"
_PATCH_VERSION = "0"

# When building releases, we can update this value on the release branch to
# reflect the current release candidate ('rc0', 'rc1') or, finally, the official
# stable release (indicated by `_VERSION_SUFFIX = ''`). Outside the context of a
# release branch, the current version is by default assumed to be a
# 'development' version, labeled 'dev'.
_VERSION_SUFFIX = "rc0"

# Example, '0.1.0-dev'
__version__ = ".".join([_MAJOR_VERSION, _MINOR_VERSION, _PATCH_VERSION])
if _VERSION_SUFFIX:
  __version__ = "{}-{}".format(__version__, _VERSION_SUFFIX)

# Required TFX version [min, max), keep depconstraint in ci.yml in sync
_INCLUSIVE_MIN_TFX_VERSION = "1.4.0"
_EXCLUSIVE_MAX_TFX_VERSION = "1.8.0"
_TFXVERSION_CONSTRAINT = (
    f">={_INCLUSIVE_MIN_TFX_VERSION},<{_EXCLUSIVE_MAX_TFX_VERSION}")

# This is a list of officially  maintained projects with their dependencies.
# Any project added here will be automatically picked up on release.
# - Key: Project name that corresponds to  folder tfx_addons.{} namespace.
# - Value: Python dependencies needed for project to work.
_PKG_METADATA = {
    # Add dependencies here for your project. Avoid using install_requires.
    "mlmd_client": [
        f"ml_pipelines_sdk{_TFXVERSION_CONSTRAINT}",
        f"ml_metadata{_TFXVERSION_CONSTRAINT}"
    ],
    "schema_curation": [
        f"tfx{_TFXVERSION_CONSTRAINT}",
    ],
    "feast_examplegen": [
        f"tfx{_TFXVERSION_CONSTRAINT}",
        "feast>=0.16.0,<1.0.0",
    ],
    "xgboost_evaluator": [
        f"tfx{_TFXVERSION_CONSTRAINT}",
        "xgboost>=1.0.0",
    ],
    "sampling": [f"tfx{_TFXVERSION_CONSTRAINT}", "tensorflow>=2.0.0"],
    "message_exit_handler": [
        f"tfx{_TFXVERSION_CONSTRAINT}",
        "kfp>=1.8,<2.0",
        "slackclient>=2.9.0,<3.0",
        "pydantic>=1.8.0,<2.0",
    ],
}



# Copyright 2021 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Sampling component definition."""

from typing import List, Optional, Text

from tfx import types
from tfx.dsl.components.base import base_beam_component, executor_spec
from tfx.types import channel_utils, standard_artifacts
from tfx.utils import json_utils

from tfx_addons.sampling.executor import Executor
from tfx_addons.sampling.spec import SamplerSpec, SamplingStrategy


class Sampler(base_beam_component.BaseBeamComponent):
  """A TFX component to sample examples.

  The sampling component wraps an Apache Beam pipeline to process
  data in an TFX pipeline. This component loads in tf.Record files from
  an earlier example artifact, processes the 'train' split by default,
  samples the split by a given label's classes, and stores the new
  set of sampled examples into its own example artifact in
  tf.Record format.

  The sampling logic uses Python's random module:
  undersampling uses random.sample, and oversampling uses
  random.choices. Support for more complex sampling algorithms may
  be added at a later date.

  By default, the component will ignore all examples with a null value
  (more precisely, a value that evaluates to False) for the given label,
  although more values can be added in as necessary. Additionally, it will
  copy all non-'train' splits, through this behavior can be changed as well.
  The component will save the examples in a user-specified number of files,
  and it can be given a name as well.

  ## Example
  ```
  # Performs transformations and feature engineering in training and serving.
  under = Sampler(
    examples=example_gen.outputs['examples'])
  ```

  Component `outputs` contains:
   - `sampled_examples`: Channel of type `standard_artifacts.Examples` for
      materialized sampled examples, based on the input splits, which includes
      copied splits unless otherwise specified by copy_others.
  """

  SPEC_CLASS = SamplerSpec
  EXECUTOR_SPEC = executor_spec.BeamExecutorSpec(Executor)

  def __init__(
      self,
      label: str,
      input_data: types.Channel = None,
      output_data: types.Channel = None,
      name: Optional[Text] = None,
      splits: Optional[List[Text]] = None,
      copy_others: Optional[bool] = True,
      shards: Optional[int] = 0,
      null_classes: Optional[List[Text]] = None,
      sampling_strategy: SamplingStrategy = SamplingStrategy.UNDERSAMPLE):
    """Construct a SamplerComponent.

    Args:
      input_data: A Channel of type `standard_artifacts.Examples`.
      output_data: A Channel of type `standard_artifacts.Examples`.
        By default, only the train split is sampled; all others are copied.
      name: Optional unique name. Necessary if multiple components are
        declared in the same pipeline.
      label: The name of the column containing class names to sample by.
      splits: A list containing splits to sample.
      copy_others: Determines whether we copy over the splits that aren't
        sampled, or just exclude them from the output artifact.
      shards: The number of files that each sampled split should
        contain. Default 0 is Beam's tfrecordio function's default.
      null_classes: A list determining which classes that we should not sample.
    """

    if not output_data:
      output_data = channel_utils.as_channel([standard_artifacts.Examples()])

    spec = SamplerSpec(
        input_data=input_data,
        output_data=output_data,
        label=label,
        name=name,
        splits=json_utils.dumps(splits),
        copy_others=int(copy_others),
        shards=shards,
        null_classes=json_utils.dumps(null_classes),
        sampling_strategy=sampling_strategy,
    )

    super().__init__(spec=spec)



# Copyright 2022 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for Logging Provider functions."""

import tensorflow as tf

from tfx_addons.message_exit_handler import constants
from tfx_addons.message_exit_handler.message_providers import logging_provider

SUCCESS_MESSAGE = """:tada: Pipeline job *test-pipeline-job* (test-project) completed successfully.

https://console.cloud.google.com/vertex-ai/locations/us-central1/pipelines/runs/test-pipeline-job"""

FAILURE_MESSAGE = """:scream: Pipeline job *test-pipeline-job* (test-project) failed.
>test error
https://console.cloud.google.com/vertex-ai/locations/us-central1/pipelines/runs/test-pipeline-job"""


class LoggingProviderTest(tf.test.TestCase):
  @staticmethod
  def get_final_status(state: str = constants.SUCCESS_STATUS,
                       error: str = "") -> str:
    """Assemble final status for tests"""
    status = {
        "state":
        state,
        "error":
        error,
        "pipelineJobResourceName":
        ("projects/test-project/locations/"
         "us-central1/pipelineJobs/test-pipeline-job"),
    }
    if error:
      status.update({"error": {"message": error}})
    return status

  def test_logging_message_provider_success(self):
    final_status = self.get_final_status()
    with self.assertLogs(level="INFO") as logs:
      message_provider = logging_provider.LoggingMessageProvider(final_status)
      message_provider.send_message()
      self.assertLen(logs.output, 1)
      self.assertEqual(
          "INFO:absl:MessageExitHandler: " + SUCCESS_MESSAGE,
          logs.output[0],
      )

  def test_logging_message_provider_failure(self):
    final_status = self.get_final_status(state=constants.FAILURE_STATUS,
                                         error="test error")
    with self.assertLogs(level="INFO") as logs:
      message_provider = logging_provider.LoggingMessageProvider(final_status)
      message_provider.send_message()
      self.assertLen(logs.output, 1)
      self.assertEqual(
          "INFO:absl:MessageExitHandler: " + FAILURE_MESSAGE,
          logs.output[0],
      )


if __name__ == "__main__":
  tf.test.main()



# Copyright 2021 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Python source which includes pipeline functions for the Credit ard fraud.
The utilities in this file are used to build a model with native Keras or with
Flax.
"""

from typing import List, Text

import tensorflow as tf
import tensorflow_model_analysis as tfma
import tensorflow_transform as tft
from tensorflow_transform.tf_metadata import schema_utils
from tfx.components.trainer.fn_args_utils import DataAccessor
from tfx_bsl.tfxio import dataset_options

_FEATURE_KEYS = [
    "V1", "V2", "V3", "V4", "V5", "V6", "V7", "V8", "V9", "V10", "V11", "V12",
    "V13", "V14", "V15", "V16", "V17", "V18", "V19", "V20", "V21", "V22",
    "V23", "V24", "V25", "V26", "V27", "V28", "Amount"
]
_LABEL_KEY = 'Class'

TRAIN_BATCH_SIZE = 20
EVAL_BATCH_SIZE = 10


def _transformed_name(key):
  return key + '_xf'


def _transformed_names(keys):
  return [_transformed_name(key) for key in keys]


# Tf.Transform considers these features as "raw"
def _get_raw_feature_spec(schema):
  return schema_utils.schema_as_feature_spec(schema).feature_spec


def _fill_in_missing(x):
  """Replace missing values in a SparseTensor.
  Fills in missing values of `x` with '' or 0, and converts to a dense tensor.
  Args:
    x: A `SparseTensor` of rank 2.  Its dense shape should have size at most 1
      in the second dimension.
  Returns:
    A rank 1 tensor where missing values of `x` have been filled in.
  """
  if not isinstance(x, tf.sparse.SparseTensor):
    return x

  default_value = '' if x.dtype == tf.string else 0
  return tf.squeeze(tf.sparse.to_dense(
      tf.SparseTensor(x.indices, x.values, [x.dense_shape[0], 1]),
      default_value),
                    axis=1)


def preprocessing_fn(inputs):
  """tf.transform's callback function for preprocessing inputs.
  Args:
    inputs: map from feature keys to raw not-yet-transformed features.
  Returns:
    Map from string feature key to transformed feature operations.
  """
  outputs = {}
  for key in _FEATURE_KEYS:
    # Preserve this feature as a dense float, setting nan's to the mean.
    outputs[_transformed_name(key)] = tft.scale_to_z_score(
        _fill_in_missing(inputs[key]))
  outputs[_transformed_name(_LABEL_KEY)] = _fill_in_missing(inputs[_LABEL_KEY])

  return outputs


def _build_estimator(config, hidden_units=None, warm_start_from=None):
  """Build an estimator for classifier fraud/not fraud
  Args:
    config: tf.estimator.RunConfig defining the runtime environment for the
      estimator (including model_dir).
    hidden_units: [int], the layer sizes of the DNN (input layer first)
    warm_start_from: Optional directory to warm start from.
  Returns:
    A dict of the following:
      - estimator: The estimator that will be used for training and eval.
      - train_spec: Spec for training.
      - eval_spec: Spec for eval.
      - eval_input_receiver_fn: Input function for eval.
  """
  real_valued_columns = [
      tf.feature_column.numeric_column(key, shape=())
      for key in _transformed_names(_FEATURE_KEYS)
  ]

  return tf.estimator.DNNLinearCombinedClassifier(
      config=config,
      dnn_feature_columns=real_valued_columns,
      dnn_hidden_units=hidden_units or [100, 70, 50, 25],
      warm_start_from=warm_start_from)


def _example_serving_receiver_fn(tf_transform_output, schema):
  """Build the serving in inputs.
  Args:
    tf_transform_output: A TFTransformOutput.
    schema: the schema of the input data.
  Returns:
    Tensorflow graph which parses examples, applying tf-transform to them.
  """
  raw_feature_spec = _get_raw_feature_spec(schema)
  raw_feature_spec.pop(_LABEL_KEY)

  raw_input_fn = tf.estimator.export.build_parsing_serving_input_receiver_fn(
      raw_feature_spec, default_batch_size=None)
  serving_input_receiver = raw_input_fn()

  transformed_features = tf_transform_output.transform_raw_features(
      serving_input_receiver.features)

  return tf.estimator.export.ServingInputReceiver(
      transformed_features, serving_input_receiver.receiver_tensors)


def _eval_input_receiver_fn(tf_transform_output, schema):
  """Build everything needed for the tf-model-analysis to run the model.
  Args:
    tf_transform_output: A TFTransformOutput.
    schema: the schema of the input data.
  Returns:
    EvalInputReceiver function, which contains:
      - Tensorflow graph which parses raw untransformed features, applies the
        tf-transform preprocessing operators.
      - Set of raw, untransformed features.
      - Label against which predictions will be compared.
  """
  # Notice that the inputs are raw features, not transformed features here.
  raw_feature_spec = _get_raw_feature_spec(schema)

  serialized_tf_example = tf.compat.v1.placeholder(dtype=tf.string,
                                                   shape=[None],
                                                   name='input_example_tensor')

  # Add a parse_example operator to the tensorflow graph, which will parse
  # raw, untransformed, tf examples.
  features = tf.io.parse_example(serialized_tf_example, raw_feature_spec)

  # Now that we have our raw examples, process them through the tf-transform
  # function computed during the preprocessing step.
  transformed_features = tf_transform_output.transform_raw_features(features)

  # The key name MUST be 'examples'.
  receiver_tensors = {'examples': serialized_tf_example}

  # NOTE: Model is driven by transformed features (since training works on the
  # materialized output of TFT, but slicing will happen on raw features.
  features.update(transformed_features)

  return tfma.export.EvalInputReceiver(
      features=features,
      receiver_tensors=receiver_tensors,
      labels=transformed_features[_transformed_name(_LABEL_KEY)])


def _input_fn(file_pattern: List[Text],
              data_accessor: DataAccessor,
              tf_transform_output: tft.TFTransformOutput,
              batch_size: int = 200) -> tf.data.Dataset:
  """Generates features and label for tuning/training.
  Args:
    file_pattern: List of paths or patterns of input tfrecord files.
    data_accessor: DataAccessor for converting input to RecordBatch.
    tf_transform_output: A TFTransformOutput.
    batch_size: representing the number of consecutive elements of returned
      dataset to combine in a single batch
  Returns:
    A dataset that contains (features, indices) tuple where features is a
      dictionary of Tensors, and indices is a single Tensor of label indices.
  """
  return data_accessor.tf_dataset_factory(
      file_pattern,
      dataset_options.TensorFlowDatasetOptions(
          batch_size=batch_size, label_key=_transformed_name(_LABEL_KEY)),
      tf_transform_output.transformed_metadata.schema)


# TFX will call this function
def trainer_fn(trainer_fn_args, schema):
  """Build the estimator using the high level API.
  Args:
    trainer_fn_args: Holds args used to train the model as name/value pairs.
    schema: Holds the schema of the training examples.
  Returns:
    A dict of the following:
      - estimator: The estimator that will be used for training and eval.
      - train_spec: Spec for training.
      - eval_spec: Spec for eval.
      - eval_input_receiver_fn: Input function for eval.
  """
  # Number of nodes in the first layer of the DNN
  first_dnn_layer_size = 100
  num_dnn_layers = 4
  dnn_decay_factor = 0.7

  train_batch_size = 40
  eval_batch_size = 40

  tf_transform_output = tft.TFTransformOutput(trainer_fn_args.transform_output)

  train_input_fn = lambda: _input_fn(  # pylint: disable=g-long-lambda
      trainer_fn_args.train_files,
      trainer_fn_args.data_accessor,
      tf_transform_output,
      batch_size=train_batch_size)

  eval_input_fn = lambda: _input_fn(  # pylint: disable=g-long-lambda
      trainer_fn_args.eval_files,
      trainer_fn_args.data_accessor,
      tf_transform_output,
      batch_size=eval_batch_size)

  train_spec = tf.estimator.TrainSpec(  # pylint: disable=g-long-lambda
      train_input_fn,
      max_steps=trainer_fn_args.train_steps)

  serving_receiver_fn = lambda: _example_serving_receiver_fn(  # pylint: disable=g-long-lambda
      tf_transform_output, schema)

  exporter = tf.estimator.FinalExporter('chicago-taxi', serving_receiver_fn)
  eval_spec = tf.estimator.EvalSpec(eval_input_fn,
                                    steps=trainer_fn_args.eval_steps,
                                    exporters=[exporter],
                                    name='credit-fraud-eval')

  # Keep multiple checkpoint files for distributed training, note that
  # keep_max_checkpoint should be greater or equal to the number of replicas to
  # avoid race condition.
  run_config = tf.estimator.RunConfig(save_checkpoints_steps=999,
                                      keep_checkpoint_max=5)

  run_config = run_config.replace(model_dir=trainer_fn_args.serving_model_dir)
  warm_start_from = trainer_fn_args.base_model

  estimator = _build_estimator(
      # Construct layers sizes with exponetial decay
      hidden_units=[
          max(2, int(first_dnn_layer_size * dnn_decay_factor**i))
          for i in range(num_dnn_layers)
      ],
      config=run_config,
      warm_start_from=warm_start_from)

  # Create an input receiver for TFMA processing
  receiver_fn = lambda: _eval_input_receiver_fn(  # pylint: disable=g-long-lambda
      tf_transform_output, schema)

  return {
      'estimator': estimator,
      'train_spec': train_spec,
      'eval_spec': eval_spec,
      'eval_input_receiver_fn': receiver_fn
  }



# Copyright 2021 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for TFX Schema Curation Custom Component."""

import tensorflow as tf
from tfx.types import channel_utils, standard_artifacts

from tfx_addons.schema_curation.component import component


class SchemaCurationTest(tf.test.TestCase):
  def testConstruct(self):
    schema_curation = component.SchemaCuration(schema=channel_utils.as_channel(
        [standard_artifacts.Schema()]), )
    self.assertEqual(standard_artifacts.Schema.TYPE_NAME,
                     schema_curation.outputs['custom_schema'].type_name)


if __name__ == '__main__':
  tf.test.main()



# Copyright 2022 The TensorFlow Authors. All Rights Reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Package Setup script for TFX Addons."""
import itertools
import os

from setuptools import find_namespace_packages, setup

PROJECT_NAME = "tfx-addons"


def get_pkg_metadata():
  # Version
  context = {}
  base_dir = os.path.dirname(os.path.abspath(__file__))
  with open(os.path.join(base_dir, "tfx_addons", "version.py")) as fp:
    exec(fp.read(), context)  # pylint: disable=exec-used

  return context["_PKG_METADATA"]


def get_version():
  # Version
  context = {}
  base_dir = os.path.dirname(os.path.abspath(__file__))
  with open(os.path.join(base_dir, "tfx_addons", "version.py")) as fp:
    exec(fp.read(), context)  # pylint: disable=exec-used

  return context["__version__"]


def get_long_description():
  base_dir = os.path.dirname(os.path.abspath(__file__))
  with open(os.path.join(base_dir, "README.md")) as fp:
    return fp.read()


TESTS_REQUIRE = ["pytest", "pylint", "pre-commit", "isort", "yapf"]

PKG_REQUIRES = get_pkg_metadata()
EXTRAS_REQUIRE = PKG_REQUIRES.copy()
EXTRAS_REQUIRE["all"] = list(
    set(itertools.chain.from_iterable(list(PKG_REQUIRES.values()))))
EXTRAS_REQUIRE["test"] = TESTS_REQUIRE

setup(
    name=PROJECT_NAME,
    version=get_version(),
    description="TFX Addons libraries",
    author="The Tensorflow Authors",
    long_description=get_long_description(),
    long_description_content_type='text/markdown',
    url="https://github.com/tensorflow/tfx-addons",
    project_urls={
        # ToDo(gcasassaez): To add docs once we have some docs integrated.
        # "Documentation": "",
        "Bug Tracker": "https://github.com/tensorflow/tfx-addons/issues",
    },
    extras_require=EXTRAS_REQUIRE,
    tests_require=TESTS_REQUIRE,
    packages=find_namespace_packages(include=[
        # Add here new library package
        "tfx_addons",
    ] + [f"tfx_addons.{m}.*"
         for m in PKG_REQUIRES] + [f"tfx_addons.{m}" for m in PKG_REQUIRES]),
    classifiers=[
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Software Development",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    python_requires=">=3.7",
    include_package_data=True,
)
